@@ -2,24 +2,17 @@ import enum
 import os
 import lark  # type: ignore
 import logging
-import json
 import pathlib
-import re
-import shutil
-import stat
 import subprocess
 
 from . import configuration
 from . import editor
 from . import qemu
-from . import image
+from . import store
 from . import utils
 from . import ssh
-from . import static
 
 from typing import (
-    cast,
-    Any,
     Sequence,
     Callable,
     List,
@@ -127,7 +120,7 @@ class RunInstruction(ImageInstruction):
     def commands(self, builder: "ImageBuilder") -> Sequence[editor.Command]:
         return [
             GuestChrootCommand(
-                self.command, builder.editor.ssh_config, builder.config.ssh_timeout
+                self.command, builder.image_editor.ssh_config, builder.config.ssh_timeout
             )
         ]
 
@@ -148,7 +141,7 @@ class CopyInstruction(ImageInstruction):
 
         def _copy_func(host_src: str) -> Callable[[], Tuple[None, None]]:
             def _inner() -> Tuple[None, None]:
-                builder.editor.copy_in(host_src, self.destination)
+                builder.image_editor.copy_in(host_src, self.destination)
                 return None, None
 
             return _inner
@@ -178,7 +171,7 @@ class AddInstruction(ImageInstruction):
 
         def _copy_func(host_src: str) -> Callable[[], Tuple[None, None]]:
             def _inner() -> Tuple[None, None]:
-                builder.editor.copy_in(host_src, self.destination)
+                builder.image_editor.copy_in(host_src, self.destination)
                 return None, None
 
             return _inner
@@ -192,7 +185,7 @@ class AddInstruction(ImageInstruction):
                 commands.append(
                     editor.GuestCommand(
                         f"bsdtar xfP - --directory={effective_destination}",
-                        builder.editor.ssh_config,
+                        builder.image_editor.ssh_config,
                         builder.config.ssh_timeout,
                         stdin=open(host_src, "rb"),
                     )
@@ -217,7 +210,7 @@ class DiskInstruction(ImageInstruction):
         return [
             editor.GuestCommand(
                 f"echo label:{self.type} | sfdisk /dev/sda",
-                builder.editor.ssh_config,
+                builder.image_editor.ssh_config,
                 builder.config.ssh_timeout,
             )
         ]
@@ -300,7 +293,7 @@ class PartitionInstruction(ImageInstruction):
         commands = [
             editor.GuestCommand(
                 f"echo '{partition_cmd}' | sfdisk /dev/sda -a",
-                builder.editor.ssh_config,
+                builder.image_editor.ssh_config,
                 builder.config.ssh_timeout,
             )
         ]
@@ -309,7 +302,7 @@ class PartitionInstruction(ImageInstruction):
             commands.append(
                 editor.GuestCommand(
                     f"mkfs.{self.format} {self.options} /dev/sda{self.number}",
-                    builder.editor.ssh_config,
+                    builder.image_editor.ssh_config,
                     builder.config.ssh_timeout,
                 )
             )
@@ -367,19 +360,21 @@ T = TypeVar("T")
 
 
 class ImageBuilder:
-    config: configuration.Config
+    config: configuration.BuildConfig
     build_dir: str
-    store: image.ImageStore
+    imgstore: store.BackendImageStore
     instructions: List[ImageInstruction]
     qemu: qemu.QemuRunner
     from_instruction: FromInstruction
     chroot_ready: bool
-    editor: editor.ImageEditor
+    image_editor: editor.ImageEditor
 
-    def __init__(self, config: configuration.Config, store: image.ImageStore) -> None:
+    def __init__(
+        self, config: configuration.BuildConfig, imgstore: store.BackendImageStore
+    ) -> None:
         self.chroot_ready = False
         self.config = config
-        self.store = store
+        self.imgstore = imgstore
 
         if config.file is None:
             imagefile_path = os.path.join(self.config.build_dir, "Imagefile")
@@ -389,7 +384,14 @@ class ImageBuilder:
         logging.info(f"Parsing Imagefile at '{imagefile_path}'")
         with open(imagefile_path, "r") as file:
             contents = file.read()
-            parsed = IMAGEFILE_PARSER(contents)
+
+            # Until https://github.com/lark-parser/lark/issues/237 is merged in lark,
+            # we can't require instructions end with a newline _or_ EOF, so we
+            # always require a newline. This leads to somewhat confusing errors if
+            # there is no newline at the end of the file.
+            #
+            # So for now, just always append a newline
+            parsed = IMAGEFILE_PARSER(contents + "\n")
             self.instructions = [
                 _build_instruction(instr) for instr in parsed.find_data("instruction")
             ]
@@ -478,16 +480,16 @@ class ImageBuilder:
     def __prepare_new_image(self) -> str:
         self.__print_step(self.from_instruction)
 
-        name = image.storage_safe_encode(self.config.name)
+        name = store.storage_safe_encode(self.config.name)
         if self.config.local is True:
             # If this is a local (non-backend) build, then use the build dir for
             # the working image as well, so we can atomically rename at the end
             working = os.path.join(self.config.build_dir, f"{name}.working")
         else:
-            working = os.path.join(self.store.working, name)
+            working = os.path.join(self.imgstore.working, name)
 
         if not self.__is_from_scratch():
-            existing = self.store.retrieve_image(self.from_instruction.source).path
+            existing = self.imgstore.retrieve_image(self.from_instruction.source).path
             logging.info(f"Copying backend file as base of new image at '{working}'")
             with open(existing, "rb") as source, open(working, "wb") as dest:
                 # Get the size of the source for the progress bar
@@ -502,7 +504,7 @@ class ImageBuilder:
 
             utils.run_check_retcode(
                 [
-                    self.store.qemu_img_bin,
+                    self.imgstore.qemu_img_bin,
                     "create",
                     "-f",
                     "qcow2",
@@ -513,8 +515,8 @@ class ImageBuilder:
         return working
 
     def __inspect_guest_chroot(self) -> None:
-        config = self.editor.ssh_config.override(
-            args=[*self.editor.ssh_config.args, "-t"]
+        config = self.image_editor.ssh_config.override(
+            args=[*self.image_editor.ssh_config.args, "-t"]
         )
         client = ssh.SshClient(config, command="unshare --fork --pid chroot /mnt /bin/sh")
 
@@ -528,7 +530,7 @@ class ImageBuilder:
 
     def __prepare_chroot(self) -> None:
         # Adapted from arch-chroot
-        self.editor.run_command_in_guest(
+        self.image_editor.run_command_in_guest(
             [
                 'mount udev "/mnt/dev" -t devtmpfs -o mode=0755,nosuid',
                 'mount proc "/mnt/proc" -t proc -o nosuid,noexec,nodev',
@@ -577,10 +579,10 @@ class ImageBuilder:
             partition_instructions
         )
         for partition_instr in ordered_partitions:
-            self.editor.run_command_in_guest(
+            self.image_editor.run_command_in_guest(
                 f"mkdir -p /mnt/{partition_instr.mount}", allowfail=True
             )
-            self.editor.run_command_in_guest(
+            self.image_editor.run_command_in_guest(
                 f"mount /dev/sda{partition_instr.number} /mnt/{partition_instr.mount}"
             )
 
@@ -601,7 +603,7 @@ class ImageBuilder:
         try:
             return GuestChrootCommand(
                 single_cmd,
-                self.editor.ssh_config,
+                self.image_editor.ssh_config,
                 self.config.ssh_timeout,
                 capture_stdout=capture_stdout,
                 capture_stderr=capture_stderr,
@@ -614,10 +616,15 @@ class ImageBuilder:
     def build(self) -> str:
         new_image = self.__prepare_new_image()
 
-        self.editor = editor.ImageEditor(self.config, new_image, self.__is_from_scratch())
+        self.image_editor = editor.ImageEditor(
+            new_image,
+            self.config.ssh_timeout,
+            self.config.qmp_timeout,
+            skip_mount=self.__is_from_scratch(),
+        )
 
         # Start the image editor
-        self.editor.edit()
+        self.image_editor.edit()
 
         # Mount the partitions to the expected locations, but don't do any other
         # mounts in the chroot (/dev, /tmp, etc) because if this is a FROM scratch
@@ -649,16 +656,16 @@ class ImageBuilder:
             for cmd in instr.commands(self):
                 cmd.run()
 
-        self.editor.close()
+        self.image_editor.close()
 
         # Everything is done. Move the built image to its destination
         if self.config.local is True:
             destination = os.path.join(self.config.build_dir, f"{self.config.name}.qcow2")
         else:
-            destination = self.store.backend_path(image.ImageSpec(self.config.name))
+            destination = self.imgstore.backend_path(store.ImageSpec(self.config.name))
 
         # Make the new image read-only before moving.
-        os.chmod(new_image, stat.S_IREAD | stat.S_IRGRP | stat.S_IROTH)
+        utils.make_path_readonly(new_image)
         os.rename(new_image, destination)
 
         return destination
